@@ -2,14 +2,21 @@
 """
 Python 2 socket module.
 """
-import time
-from gevent import _socketcommon
-from gevent.hub import PYPY
+from __future__ import absolute_import
+from __future__ import print_function
 
-for key in _socketcommon.__dict__:
-    if key.startswith('__') or key in _socketcommon.__py3_imports__ or key in _socketcommon.__extensions__:
-        continue
-    globals()[key] = getattr(_socketcommon, key)
+# Our import magic sadly makes this warning useless
+# pylint: disable=undefined-variable
+import sys
+
+from gevent import _socketcommon
+from gevent._util import copy_globals
+from gevent._compat import PYPY
+from gevent.timeout import Timeout
+
+copy_globals(_socketcommon, globals(),
+             names_to_ignore=_socketcommon.__py3_imports__ + _socketcommon.__extensions__,
+             dunder_names_to_keep=())
 
 __socket__ = _socketcommon.__socket__
 __implements__ = _socketcommon._implements
@@ -31,51 +38,57 @@ except AttributeError:
                       'gettimeout', 'shutdown')
 else:
     # Python 2 doesn't natively support with statements on _fileobject;
-    # but it eases our test cases if we can do the same with on both Py3
-    # and Py2. Implementation copied from Python 3
-    if not hasattr(_fileobject, '__enter__'):
-        # we could either patch in place:
-        #_fileobject.__enter__ = lambda self: self
-        #_fileobject.__exit__ = lambda self, *args: self.close() if not self.closed else None
-        # or we could subclass. subclassing has the benefit of not
-        # changing the behaviour of the stdlib if we're just imported; OTOH,
-        # under Python 2.6/2.7, test_urllib2net.py asserts that the class IS
-        # socket._fileobject (sigh), so we have to work around that.
-        class _fileobject(_fileobject):
+    # but it substantially eases our test cases if we can do the same with on both Py3
+    # and Py2. (For this same reason we make the socket itself a context manager.)
+    # Implementation copied from Python 3
+    assert not hasattr(_fileobject, '__enter__')
+    # we could either patch in place:
+    #_fileobject.__enter__ = lambda self: self
+    #_fileobject.__exit__ = lambda self, *args: self.close() if not self.closed else None
+    # or we could subclass. subclassing has the benefit of not
+    # changing the behaviour of the stdlib if we're just imported; OTOH,
+    # under Python 2.6/2.7, test_urllib2net.py asserts that the class IS
+    # socket._fileobject (sigh), so we have to work around that.
 
-            def __enter__(self):
-                return self
+    # We also make it call our custom socket closing method that disposes
+    # of IO watchers but not the actual socket itself.
 
-            def __exit__(self, *args):
-                if not self.closed:
-                    self.close()
+    # Python 2 relies on reference counting to close sockets, so this is all
+    # very ugly and fragile.
 
-if sys.version_info[:2] < (2, 7):
-    _get_memory = buffer
-else:
-    def _get_memory(data):
-        try:
-            mv = memoryview(data)
-            if mv.shape:
-                return mv
-            # No shape, probably working with a ctypes object,
-            # or something else exotic that supports the buffer interface
-            return mv.tobytes()
-        except TypeError:
-            # fixes "python2.7 array.array doesn't support memoryview used in
-            # gevent.socket.send" issue
-            # (http://code.google.com/p/gevent/issues/detail?id=94)
-            return buffer(data)
+    class _fileobject(_fileobject): # pylint:disable=function-redefined
+        __slots__ = (
+            '__weakref__',
+        )
 
+        def __enter__(self):
+            return self
+
+        def __exit__(self, *args):
+            if not self.closed:
+                self.close()
+
+        def close(self):
+            if self._sock is not None:
+                self._sock._drop_events_and_close(closefd=False)
+            super(_fileobject, self).close()
+
+
+from gevent._greenlet_primitives import get_memory as _get_memory
 
 class _closedsocket(object):
-    __slots__ = []
+    __slots__ = ()
 
-    def _dummy(*args, **kwargs):
+    def _dummy(*args, **kwargs): # pylint:disable=no-method-argument,unused-argument
         raise error(EBADF, 'Bad file descriptor')
     # All _delegate_methods must also be initialized here.
     send = recv = recv_into = sendto = recvfrom = recvfrom_into = _dummy
 
+    def __nonzero__(self):
+        return False
+
+    __bool__ = __nonzero__
+
     if PYPY:
 
         def _drop(self):
@@ -88,9 +101,11 @@ class _closedsocket(object):
 
 
 timeout_default = object()
+gtype = type
 
+from gevent._hub_primitives import wait_on_socket as _wait_on_socket
 
-class socket(object):
+class socket(_socketcommon.SocketMixin):
     """
     gevent `socket.socket <https://docs.python.org/2/library/socket.html#socket-objects>`_
     for Python 2.
@@ -98,23 +113,33 @@ class socket(object):
     This object should have the same API as the standard library socket linked to above. Not all
     methods are specifically documented here; when they are they may point out a difference
     to be aware of or may document a method the standard library does not.
+
+    .. versionchanged:: 1.5.0
+        This object is a context manager, returning itself, like in Python 3.
     """
 
+    # pylint:disable=too-many-public-methods
+
+    # TODO: Define __slots__.
+
     def __init__(self, family=AF_INET, type=SOCK_STREAM, proto=0, _sock=None):
+        timeout = _socket.getdefaulttimeout()
         if _sock is None:
             self._sock = _realsocket(family, type, proto)
-            self.timeout = _socket.getdefaulttimeout()
         else:
             if hasattr(_sock, '_sock'):
-                self._sock = _sock._sock
-                self.timeout = getattr(_sock, 'timeout', False)
-                if self.timeout is False:
-                    self.timeout = _socket.getdefaulttimeout()
-            else:
-                self._sock = _sock
-                self.timeout = _socket.getdefaulttimeout()
+                timeout = getattr(_sock, 'timeout', timeout)
+                while hasattr(_sock, '_sock'):
+                    # passed a gevent socket or a native
+                    # socket._socketobject. Unwrap this all the way to the
+                    # native _socket.socket.
+                    _sock = _sock._sock
+
+            self._sock = _sock
+
             if PYPY:
                 self._sock._reuse()
+        self.timeout = timeout
         self._sock.setblocking(0)
         fileno = self._sock.fileno()
         self.hub = get_hub()
@@ -122,6 +147,12 @@ class socket(object):
         self._read_event = io(fileno, 1)
         self._write_event = io(fileno, 2)
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, t, v, tb):
+        self.close()
+
     def __repr__(self):
         return '<%s at %s %s>' % (type(self).__name__, hex(id(self)), self._formatinfo())
 
@@ -129,6 +160,7 @@ class socket(object):
         return '<%s %s>' % (type(self).__name__, self._formatinfo())
 
     def _formatinfo(self):
+        # pylint:disable=broad-except
         try:
             fileno = self.fileno()
         except Exception as ex:
@@ -161,31 +193,12 @@ class socket(object):
 
     ref = property(_get_ref, _set_ref)
 
-    def _wait(self, watcher, timeout_exc=timeout('timed out')):
-        """Block the current greenlet until *watcher* has pending events.
-
-        If *timeout* is non-negative, then *timeout_exc* is raised after *timeout* second has passed.
-        By default *timeout_exc* is ``socket.timeout('timed out')``.
-
-        If :func:`cancel_wait` is called, raise ``socket.error(EBADF, 'File descriptor was closed in another greenlet')``.
-        """
-        if watcher.callback is not None:
-            raise _socketcommon.ConcurrentObjectUseError('This socket is already used by another greenlet: %r' % (watcher.callback, ))
-        if self.timeout is not None:
-            timeout = Timeout.start_new(self.timeout, timeout_exc, ref=False)
-        else:
-            timeout = None
-        try:
-            self.hub.wait(watcher)
-        finally:
-            if timeout is not None:
-                timeout.cancel()
+    _wait = _wait_on_socket
 
     def accept(self):
-        sock = self._sock
-        while True:
+        while 1:
             try:
-                client_socket, address = sock.accept()
+                client_socket, address = self._sock.accept()
                 break
             except error as ex:
                 if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
@@ -197,14 +210,41 @@ class socket(object):
             client_socket._drop()
         return sockobj, address
 
-    def close(self, _closedsocket=_closedsocket, cancel_wait_ex=cancel_wait_ex):
-        # This function should not reference any globals. See Python issue #808164.
-        self.hub.cancel_wait(self._read_event, cancel_wait_ex)
-        self.hub.cancel_wait(self._write_event, cancel_wait_ex)
-        s = self._sock
-        self._sock = _closedsocket()
+
+    def _drop_ref_on_close(self, sock):
+        # See the same method in _socket3.py. We just can't be as deterministic
+        # as we can on Python 3.
+        scheduled_new = self.hub.loop.closing_fd(sock.fileno())
         if PYPY:
-            s._drop()
+            meth = sock._drop
+        else:
+            meth = sock.fileno # Still keep it alive if we need to
+        if scheduled_new:
+            self.hub.loop.run_callback(meth)
+        else:
+            meth()
+
+    def close(self, _closedsocket=_closedsocket):
+        if not self._sock:
+            return
+
+        # This function should not reference any globals. See Python issue #808164.
+
+        # First, break any reference to the loop.io objects. Our
+        # fileno, which they were tied to, is about to be free to be
+        # reused, so these objects are no longer functional.
+        self._drop_events_and_close()
+
+        # Next, change self._sock. On CPython, this drops a
+        # reference, and if it was the last reference, __del__ will
+        # close it. (We cannot close it, makefile() relies on
+        # reference counting like this, and it may be shared among
+        # multiple wrapper objects). Methods *must not* cache
+        # `self._sock` separately from
+        # self._write_event/self._read_event, or they will be out of
+        # sync and we may get inappropriate errors. (See
+        # test__hub:TestCloseSocketWhilePolling for an example).
+        self._sock = _closedsocket()
 
     @property
     def closed(self):
@@ -213,29 +253,24 @@ class socket(object):
     def connect(self, address):
         if self.timeout == 0.0:
             return self._sock.connect(address)
-        sock = self._sock
-        if isinstance(address, tuple):
-            r = getaddrinfo(address[0], address[1], sock.family, sock.type, sock.proto)
-            address = r[0][-1]
-        if self.timeout is not None:
-            timer = Timeout.start_new(self.timeout, timeout('timed out'))
-        else:
-            timer = None
+
+        address = _socketcommon._resolve_addr(self._sock, address)
+
+        timer = Timeout._start_new_or_dummy(self.timeout, timeout('timed out'))
         try:
-            while True:
-                err = sock.getsockopt(SOL_SOCKET, SO_ERROR)
+            while 1:
+                err = self._sock.getsockopt(SOL_SOCKET, SO_ERROR)
                 if err:
                     raise error(err, strerror(err))
-                result = sock.connect_ex(address)
+                result = self._sock.connect_ex(address)
                 if not result or result == EISCONN:
                     break
-                elif (result in (EWOULDBLOCK, EINPROGRESS, EALREADY)) or (result == EINVAL and is_windows):
+                if (result in (EWOULDBLOCK, EINPROGRESS, EALREADY)) or (result == EINVAL and is_windows):
                     self._wait(self._write_event)
                 else:
                     raise error(result, strerror(result))
         finally:
-            if timer is not None:
-                timer.cancel()
+            timer.close()
 
     def connect_ex(self, address):
         try:
@@ -243,10 +278,9 @@ class socket(object):
         except timeout:
             return EAGAIN
         except error as ex:
-            if type(ex) is error:
+            if type(ex) is error: # pylint:disable=unidiomatic-typecheck
                 return ex.args[0]
-            else:
-                raise  # gaierror is not silented by connect_ex
+            raise  # gaierror is not silenced by connect_ex
 
     def dup(self):
         """dup() -> socket object
@@ -258,7 +292,14 @@ class socket(object):
     def makefile(self, mode='r', bufsize=-1):
         # Two things to look out for:
         # 1) Closing the original socket object should not close the
-        #    socket (hence creating a new instance)
+        #    fileobject (hence creating a new socket instance);
+        #    An alternate approach is what _socket3.py does, which is to
+        #    keep count of the times makefile objects have been opened (Py3's
+        #    SocketIO helps with that). But the newly created socket, which
+        #    has its own read/write watchers, does need those to be closed
+        #    when the fileobject is; our custom subclass does that. Note that
+        #    we can't pass the 'close=True' argument, as that causes reference counts
+        #    to get screwed up, and Python2 sockets rely on those.
         # 2) The resulting fileobject must keep the timeout in order
         #    to be compatible with the stdlib's socket.makefile.
         # Pass self as _sock to preserve timeout.
@@ -268,10 +309,9 @@ class socket(object):
         return fobj
 
     def recv(self, *args):
-        sock = self._sock  # keeping the reference so that fd is not closed during waiting
-        while True:
+        while 1:
             try:
-                return sock.recv(*args)
+                return self._sock.recv(*args)
             except error as ex:
                 if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
                     raise
@@ -280,10 +320,9 @@ class socket(object):
             self._wait(self._read_event)
 
     def recvfrom(self, *args):
-        sock = self._sock
-        while True:
+        while 1:
             try:
-                return sock.recvfrom(*args)
+                return self._sock.recvfrom(*args)
             except error as ex:
                 if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
                     raise
@@ -291,10 +330,9 @@ class socket(object):
             self._wait(self._read_event)
 
     def recvfrom_into(self, *args):
-        sock = self._sock
-        while True:
+        while 1:
             try:
-                return sock.recvfrom_into(*args)
+                return self._sock.recvfrom_into(*args)
             except error as ex:
                 if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
                     raise
@@ -302,10 +340,9 @@ class socket(object):
             self._wait(self._read_event)
 
     def recv_into(self, *args):
-        sock = self._sock
-        while True:
+        while 1:
             try:
-                return sock.recv_into(*args)
+                return self._sock.recv_into(*args)
             except error as ex:
                 if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
                     raise
@@ -313,116 +350,40 @@ class socket(object):
             self._wait(self._read_event)
 
     def send(self, data, flags=0, timeout=timeout_default):
-        sock = self._sock
         if timeout is timeout_default:
             timeout = self.timeout
         try:
-            return sock.send(data, flags)
+            return self._sock.send(data, flags)
         except error as ex:
-            if ex.args[0] != EWOULDBLOCK or timeout == 0.0:
+            if ex.args[0] not in _socketcommon.GSENDAGAIN or timeout == 0.0:
                 raise
             sys.exc_clear()
             self._wait(self._write_event)
             try:
-                return sock.send(data, flags)
+                return self._sock.send(data, flags)
             except error as ex2:
                 if ex2.args[0] == EWOULDBLOCK:
                     return 0
                 raise
 
-    def __send_chunk(self, data_memory, flags, timeleft, end):
-        """
-        Send the complete contents of ``data_memory`` before returning.
-        This is the core loop around :meth:`send`.
-
-        :param timeleft: Either ``None`` if there is no timeout involved,
-           or a float indicating the timeout to use.
-        :param end: Either ``None`` if there is no timeout involved, or
-           a float giving the absolute end time.
-        :return: An updated value for ``timeleft`` (or None)
-        :raises timeout: If ``timeleft`` was given and elapsed while
-           sending this chunk.
-        """
-        data_sent = 0
-        len_data_memory = len(data_memory)
-        while data_sent < len_data_memory:
-            chunk = data_memory[data_sent:]
-            if timeleft is None:
-                data_sent += self.send(chunk, flags)
-            elif timeleft <= 0:
-                # Check before sending to guarantee a check
-                # happens even if each chunk successfully sends its data
-                # (especially important for SSL sockets since they have large
-                # buffers)
-                raise timeout('timed out')
-            else:
-                data_sent += self.send(chunk, flags, timeout=timeleft)
-                timeleft = end - time.time()
-
-        return timeleft
-
     def sendall(self, data, flags=0):
         if isinstance(data, unicode):
             data = data.encode()
         # this sendall is also reused by gevent.ssl.SSLSocket subclass,
         # so it should not call self._sock methods directly
         data_memory = _get_memory(data)
-        len_data_memory = len(data_memory)
-        if not len_data_memory:
-            # Don't send empty data, can cause SSL EOFError.
-            # See issue 719
-            return 0
-
-        # On PyPy up through 2.6.0, subviews of a memoryview() object
-        # copy the underlying bytes the first time the builtin
-        # socket.send() method is called. On a non-blocking socket
-        # (that thus calls socket.send() many times) with a large
-        # input, this results in many repeated copies of an ever
-        # smaller string, depending on the networking buffering. For
-        # example, if each send() can process 1MB of a 50MB input, and
-        # we naively pass the entire remaining subview each time, we'd
-        # copy 49MB, 48MB, 47MB, etc, thus completely killing
-        # performance. To workaround this problem, we work in
-        # reasonable, fixed-size chunks. This results in a 10x
-        # improvement to bench_sendall.py, while having no measurable impact on
-        # CPython (since it doesn't copy at all the only extra overhead is
-        # a few python function calls, which is negligible for large inputs).
-
-        # See https://bitbucket.org/pypy/pypy/issues/2091/non-blocking-socketsend-slow-gevent
-
-        # Too small of a chunk (the socket's buf size is usually too
-        # small) results in reduced perf due to *too many* calls to send and too many
-        # small copies. With a buffer of 143K (the default on my system), for
-        # example, bench_sendall.py yields ~264MB/s, while using 1MB yields
-        # ~653MB/s (matching CPython). 1MB is arbitrary and might be better
-        # chosen, say, to match a page size?
-        chunk_size = max(self.getsockopt(SOL_SOCKET, SO_SNDBUF), 1024 * 1024)
-
-        data_sent = 0
-        end = None
-        timeleft = None
-        if self.timeout is not None:
-            timeleft = self.timeout
-            end = time.time() + timeleft
-
-        while data_sent < len_data_memory:
-            chunk_end = min(data_sent + chunk_size, len_data_memory)
-            chunk = data_memory[data_sent:chunk_end]
-
-            timeleft = self.__send_chunk(chunk, flags, timeleft, end)
-            data_sent += len(chunk) # Guaranteed it sent the whole thing
+        return _socketcommon._sendall(self, data_memory, flags)
 
     def sendto(self, *args):
-        sock = self._sock
         try:
-            return sock.sendto(*args)
+            return self._sock.sendto(*args)
         except error as ex:
             if ex.args[0] != EWOULDBLOCK or self.timeout == 0.0:
                 raise
             sys.exc_clear()
             self._wait(self._write_event)
             try:
-                return sock.sendto(*args)
+                return self._sock.sendto(*args)
             except error as ex2:
                 if ex2.args[0] == EWOULDBLOCK:
                     return 0
@@ -462,10 +423,19 @@ class socket(object):
     type = property(lambda self: self._sock.type)
     proto = property(lambda self: self._sock.proto)
 
+    def fileno(self):
+        return self._sock.fileno()
+
+    def getsockname(self):
+        return self._sock.getsockname()
+
+    def getpeername(self):
+        return self._sock.getpeername()
+
     # delegate the functions that we haven't implemented to the real socket object
 
     _s = "def %s(self, *args): return self._sock.%s(*args)\n\n"
-
+    _m = None
     for _m in set(_socketmethods) - set(locals()):
         exec(_s % (_m, _m,))
     del _m, _s
@@ -483,8 +453,9 @@ SocketType = socket
 
 if hasattr(_socket, 'socketpair'):
 
-    def socketpair(*args):
-        one, two = _socket.socketpair(*args)
+    def socketpair(family=getattr(_socket, 'AF_UNIX', _socket.AF_INET),
+                   type=_socket.SOCK_STREAM, proto=0):
+        one, two = _socket.socketpair(family, type, proto)
         result = socket(_sock=one), socket(_sock=two)
         if PYPY:
             one._drop()
@@ -495,8 +466,8 @@ elif 'socketpair' in __implements__:
 
 if hasattr(_socket, 'fromfd'):
 
-    def fromfd(*args):
-        s = _socket.fromfd(*args)
+    def fromfd(fd, family, type, proto=0):
+        s = _socket.fromfd(fd, family, type, proto)
         result = socket(_sock=s)
         if PYPY:
             s._drop()
@@ -510,6 +481,8 @@ if hasattr(__socket__, 'ssl'):
     def ssl(sock, keyfile=None, certfile=None):
         # deprecated in 2.7.9 but still present;
         # sometimes backported by distros. See ssl.py
+        # Note that we import gevent.ssl, not _ssl2, to get the correct
+        # version.
         from gevent import ssl as _sslmod
         # wrap_socket is 2.7.9/backport, sslwrap_simple is older. They take
         # the same arguments.
